import operator

from django.test import TestCase

from django_filtering.filter import Filter, CustomOperators


class TestFilterMapping(TestCase):
    def test_json(self):
        this_filter = Filter(
            path="a.b.c",
            operator="lt",
            value="test"
        )

        self.assertDictEqual(this_filter.json, {"a__b__c__lt": "test"})

    def test_invalid_operator(self):
        self.assertRaises(
            ValueError,
            Filter,
            path="a",
            operator="invalid operator test",  # noqa
            value="test",
        )

    def test_mapped_operator(self):
        lt = Filter(
            path="a",
            operator="lt",
            value=1
        )

        regex = Filter(
            path="a",
            operator="regex",
            value=r"^test$",
        )

        self.assertEqual(lt.mapped_operator, operator.lt)
        self.assertEqual(regex.mapped_operator, CustomOperators.regex)

    def test_merge_to_dict(self):
        filters = [
            Filter(
                path=f"test.{i}",
                operator="exact",
                value="test",
            ) for i in range(0, 5)
        ]

        json_filters = {'test__0__exact': 'test', 'test__1__exact': 'test', 'test__2__exact': 'test', 'test__3__exact': 'test', 'test__4__exact': 'test'}

        self.assertDictEqual(json_filters, Filter.merge_to_dict(*filters))



from django.test import TestCase

from .. import FilterSet, Filter
from ..models import Requirement, Job, Person


class TestFiltering(TestCase):
    def setUp(self):
        self.requirements = Requirement.objects.bulk_create([
            Requirement(
                details=f"REQ{i}"
            ) for i in range(0, 3)
        ])

        self.jobs = Job.objects.bulk_create([
            Job(
                name=f"JOB{i}",
                value=i,
                requirement=self.requirements[i],
            ) for i in range(0, 3)
        ])

        self.people = Person.objects.bulk_create([
            Person(
                name=name,
                value=idx,
                job=self.jobs[idx]
            ) for idx, name in enumerate(["Joe Mama", "Ben Bunger", "Larry"])
        ])

    def test_filtering(self):
        filter_1 = Filter(
            path="name",
            operator="exact",
            value="Joe Mama"
        )
        filter_2 = Filter(
            path="job.name",
            operator="exact",
            value="JOB1"
        )
        filter_3 = Filter(
            path="job.requirement.details",
            operator="exact",
            value="REQ2"
        )
        filter_4 = Filter(
            path="name",
            operator="contains",
            value="TEST BAD DATA",
        )
        filter_5 = Filter(
            path="job.value",
            operator="gte",
            value="1",
        )

        a = FilterSet(filters=[filter_1]).filter(Person)
        b = FilterSet(filters=[filter_2]).filter(Person)
        c = FilterSet(filters=[filter_3]).filter(Person)
        d = FilterSet(filters=[filter_4]).filter(Person)
        e = FilterSet(filters=[filter_5]).filter(Person)

        self.assertQuerySetEqual(a, [Person.objects.get(pk=1)])
        self.assertQuerySetEqual(b, [Person.objects.get(pk=2)])
        self.assertQuerySetEqual(c, [Person.objects.get(pk=3)])
        self.assertQuerySetEqual(d, [])
        self.assertQuerySetEqual(list(e), [Person.objects.get(pk=2), Person.objects.get(pk=3)])

    def test_instance_filtering(self):
        people = list(Person.objects.all())

        a = FilterSet(filters=[
            Filter(
                path="name",
                operator="exact",
                value="Joe Mama"
            )
        ]).filter(people)

        b = FilterSet(filters=[
            Filter(
                path="job.name",
                operator="in",
                value=[
                    "JOB1",
                    "JOB2",
                ]
            )
        ]).filter(people)

        c = FilterSet(filters=[
            Filter(
                path="job.requirement.details",
                operator="regex",
                value=r"^REQ(?:1|2)$"
            )
        ]).filter(people)

        d = FilterSet(filters=[
            Filter(
                path="name",
                operator="istartswith",
                value="lar"
            )
        ]).filter(people)

        e = FilterSet(filters=[
            Filter(
                path="name",
                operator="endswith",
                value="ry"
            )
        ]).filter(people)

        f = FilterSet(filters=[
            Filter(
                path="name",
                operator="isnotnull",
                value=None
            )
        ]).filter(people)

        g = FilterSet(filters=[
            Filter(
                path="name",
                operator="isnull",
                value=None
            )
        ]).filter(people)

        self.assertEqual(len(a), 1)
        self.assertEqual(a[0], self.people[0])

        self.assertEqual(len(b), 2)
        self.assertEqual(b[0], self.people[1])
        self.assertEqual(b[1], self.people[2])

        self.assertEqual(len(c), 2)
        self.assertEqual(c[0], self.people[1])
        self.assertEqual(c[1], self.people[2])

        self.assertEqual(len(d), 1)
        self.assertEqual(d[0], self.people[2])
        self.assertEqual(e[0], d[0])

        self.assertEqual(len(f), 3)
        self.assertEqual(len(g), 0)



from django.db import models


class Requirement(models.Model):
    details = models.CharField(max_length=100)


class Job(models.Model):
    name = models.CharField(max_length=100)
    value = models.IntegerField()
    requirement = models.ForeignKey(Requirement, on_delete=models.CASCADE)


class Person(models.Model):
    name = models.CharField(max_length=100)
    value = models.IntegerField()
    job = models.ForeignKey(Job, on_delete=models.CASCADE)



SECRET_KEY = 'django-insecure-bs_wsbkt2&e$-+lver%rhjd6@#13_6t_x1&$x7g9n%^yu*#fbb'

DEBUG = True

INSTALLED_APPS = [
    'django_filtering',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}



import functools
import operator
import re
from dataclasses import dataclass
from datetime import datetime, date
from functools import reduce
from typing import Literal, get_args, Union, Type, TypeVar, overload, Iterable, Callable

from django.db import models
from django.db.models.query import QuerySet

Value = str | int | float | datetime | date | None
T = TypeVar("T", bound=models.Model)

# TODO: implement Q based querying for inverse lookups
Lookup = Literal[
    "exact",
    "iexact",
    "contains",
    "icontains",
    "in",
    "gt",
    "gte",
    "lt",
    "lte",
    "startswith",
    "istartswith",
    "endswith",
    "iendswith",
    "range",
    "date",
    "year",
    "iso_year",
    "month",
    "day",
    "week",
    "week_day",
    "iso_week_day",
    "quarter",
    "time",
    "hour",
    "minute",
    "second",
    "isnull",
    "isnotnull",  # not a Django filter fyi
    "regex",
    "iregex",
]

class CustomOperators:
    @staticmethod
    def startswith(value, _str):
        return value.startswith(_str)

    @staticmethod
    def istartswith(value, _str):
        return CustomOperators.startswith(value.lower(), _str.lower())

    @staticmethod
    def endswith(value, _str):
        return value.endswith(_str)

    @staticmethod
    def iendswith(value, _str):
        return CustomOperators.endswith(value.lower(), _str.lower())

    @staticmethod
    def range(value, a, b):
        return a < value < b

    @staticmethod
    def isnull(value, *_):
        return value is None

    @staticmethod
    def isnotnull(value, *_):
        return value is not None

    @staticmethod
    def regex(value, regex):
        return bool(re.search(regex, value))

    @staticmethod
    def iregex(value, regex):
        return bool(re.search(regex, value, flags=re.IGNORECASE))

    @staticmethod
    def contains(value, comparator):
        # Custom operator because strings need to be on the left operand
        # and operator.contains is reversed.
        if isinstance(value, str):
            return value in comparator
        return operator.contains(value, comparator)

class OperatorMap:
    _operator_mapping = {
        "exact": operator.eq,
        "iexact": operator.eq,
        "contains": CustomOperators.contains,
        "in": CustomOperators.contains,
        "gt": operator.gt,
        "gte": operator.ge,
        "lt": operator.lt,
        "lte": operator.le,
        "startswith": CustomOperators.startswith,
        "istartswith": CustomOperators.istartswith,
        "endswith": CustomOperators.endswith,
        "iendswith": CustomOperators.iendswith,
        "range": CustomOperators.range,
        "isnull": CustomOperators.isnull,
        "isnotnull": CustomOperators.isnotnull,
        "regex": CustomOperators.regex,
        "iregex": CustomOperators.iregex,
    }

    def __getitem__(self, item):
        if item in self._operator_mapping:
            return self._operator_mapping[item]
        else:
            raise NotImplementedError(item)


@dataclass
class Filter:
    path: str
    operator: Lookup
    value: Value | list[Value]

    def __post_init__(self):
        if self.operator not in get_args(Lookup):
            raise ValueError(f"Invalid Django lookup: {self.operator}")

    @property
    def json(self) -> dict:
        return {f"{self.path.replace('.', '__')}__{self.operator}": self.value}

    @staticmethod
    def merge_to_dict(*filters: Union["Filter", list["Filter"]]) -> dict:
        return reduce(operator.ior, [i.json for sub in filters for i in (sub if isinstance(sub, list) else [sub])], {})

    @staticmethod
    def from_list(*filter_dicts: dict | list[dict]) -> list["Filter"]:
        return [Filter(**_filter) for sub in filter_dicts for _filter in (sub if isinstance(sub, list) else [sub])]

    @property
    def mapped_operator(self) -> Callable:
        return OperatorMap()[self.operator]

    def resolve_attr(self, obj: T):
        return functools.reduce(lambda d, key: getattr(d, key), self.path.split("."), obj)


@dataclass
class FilterSet:
    filters: list[Filter]

    @overload
    def filter(self, model: Type[T]) -> QuerySet[T]:
        """
        Execute a Django queryset against supplied model type.

        :param model: The Django model to query against.
        :type model: Type[models.Model]
        :return: The queryset result.
        """
        ...

    @overload
    def filter(self, model: Iterable[T]) -> list[T]:
        """
        Filter a list of instance objects.

        :param model: The list of objects to filter against.
        :type model: Iterable[models.Model]
        :return: A filtered list of the supplied instance objects.
        """
        ...

    def filter(self, model: Type[T] | Iterable[T]) -> QuerySet[T] | list[T]:
        results = []
        if isinstance(model, list | Iterable):
            for instance in model:
                for _filter in self.filters:
                    value = []
                    if not isinstance(_filter.value, list):
                        value = [_filter.value]
                    else:
                        value.append(_filter.value)

                    if _filter.mapped_operator(*[_filter.resolve_attr(instance), *value]):
                        results.append(instance)
                        continue
            return results
        else:
            if issubclass(model, models.Model):
                return model.objects.filter(**Filter.merge_to_dict(self.filters))
            raise ValueError(f"Invalid Django model type: {model}")




from django.db import models
from functools import reduce
import operator

def merge(*args: str | list[str]) -> dict:
	return reduce(operator.ior, [{i: "1"} for sub in args for i in (sub if isinstance(sub, list) else [sub])], {})

print(merge(["a", "b", "c"]))
print(merge("a", "b", "c"))
print(merge("a", "b", "c", ["d", "e", "f"]))


from django_filtering import Filter, models

my_filters = [
    {
        "path": "job.name",
        "operator": "iexact",
        "value": "Engineer"
    },
    {
        "path": "job.requirement.name",
        "operator": "icontains",
        "value": "Plumbing"
    },
]

print(Filter.merge_to_dict(Filter.from_list(my_filters)))